/* apache includes */
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_conf_globals.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_main.h"
#include "http_core.h"
#include "http_log.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>


typedef struct _INPUT {
  char * name;
  char * label;
  char * value;

  char * (*f_check)(char *); 
  char * error;

  char * type;
  char * size;
  char * max_len; 

  /* used only for type == SELECT... */
  /* if option_list!=NULL -> INPUT is a "<SELECT>" */
  char ** option_list;	    /* MUST BE A 'NULL' TERMINATED ARRAY */
} INPUT;


typedef struct _FORM {
  INPUT* * *inputs;
 
  #ifndef CGI_POST 
  #define CGI_POST 0
  #endif
  #ifndef CGI_GET
  #define CGI_GET 1
  #endif
  int method;

  char * action;
} FORM;

#define CGI_POST 0
#define CGI_GET  1
#define QOS_CREATE_ACTION "https://qossun.citi.umich.edu/create_reservation.html"
#define QOS_CANCEL_ACTION "https://qossun.citi.umich.edu/cancel_reservation.html"

#define QOS_CREATE_HEADER \
"<HTML><HEAD><TITLE>Authenticated QoS Reservation</TITLE></HEAD>\
<BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\">\
<H3>Authenticated QoS Create Reservation</H3>\
<P>This reservation interface to \
<A HREF=\"http://www-fp.mcs.anl.gov/qos/\">GARA</A>\
 is protected by mutual SSL authentication.<P>\
<P><HR><P><UL><P>\n"                                                            

#define QOS_CANCEL_HEADER \
"<HTML><HEAD><TITLE>Authenticated QoS Reservation</TITLE></HEAD>\
<BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\">\
<H3>Authenticated QoS Cancel Reservation</H3>\
<P>This reservation interface to \
<A HREF=\"http://www-fp.mcs.anl.gov/qos/\">GARA</A>\
 is protected by mutual SSL authentication.<P>\
<P><HR><P><UL><P>\n"

//#define QOS_FOOTER \
//"<P></UL><HR>page maintained by: <A HREF=\"mailto:someone@citi.umich.edu\">someone@citi.umich.edu</A></BODY></HTML>\n"

#define QOS_FOOTER "</BODY></HTML>\n"

int cgi_init_post(request_rec *r, char ** POST);
char * cgi_var(char *name, int type, char **POST, pool *p);
char * cgi_get_var(char *name, char **post, pool *p);  
char * fix_query(char * query);
char * int_check(char * val);
char * inet_addr_check(char * value);
char * check_inet_4_addr(char * value);
INPUT * INPUT_init(char * name, char * label, char* (*f_check)(char *), char * type, char * size, char *max_len, char ** option_list, pool *);
FORM * FORM_init(INPUT *** inputs, int method, char * action, pool *p);
void FORM_print(FORM * f, request_rec *r);
void INPUT_add_error( INPUT * , char *, pool *);
void INPUT_print( INPUT * input, request_rec *r );
char * FORM_find_value(FORM *, char * );
void FORM_error( FORM * form, char* input_name, char * err_msg, pool * );
int post_form(request_rec *);
int read_form(request_rec *);
int setvar(FORM * f, char *post, request_rec *r);
FORM * set_up_form(pool *);
char *send_var( FORM * f, request_rec *r );
int set_date( FORM * form, request_rec *r );
char * find_offset(FORM * form, request_rec *r);
int sec_offset(time_t * now, int sec, int min, int hour, int day, int month, 
	       int year);
char * copy_string(pool *p, char * src);
int compare_strings(char * one, char * two);
void print_file( char * filename );


char * monthtt[13] =
{ "JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC", NULL};

static int e2eform_handler(request_rec *r) {
  char *data = NULL;

  if(!strncmp(&(r->uri[1]), "create_reservation.html", 16)) {
    if(!strncmp(r->method, "GET", 3)) {
#ifdef DEBUG
      fprintf(stderr, "recieved a GET request\n");
#endif
      if(!post_form(r))
        return OK;
      else
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    else if(!strncmp(r->method, "POST", 4)){
#ifdef DEBUG
      fprintf(stderr, "received a new POST request\n");
#endif
      if(!read_form(r))
        return DECLINED;
      else
        return HTTP_INTERNAL_SERVER_ERROR;
    }
  }
  else if(!strncmp(&r->uri[1], "cancel_reservation.html", 23)) {
    if(!strncmp(r->method, "GET", 3)) {
      fprintf(stderr, "cancel get request\n");
      r->content_type = "text/html";
      ap_send_http_header(r);

      ap_rputs(QOS_CANCEL_HEADER, r);
      ap_rprintf(r, "<FORM METHOD=\"POST\" ACTION=\"%s\">\n", 
               "https://qossun.citi.umich.edu/cancel_reservation.html");
      ap_rputs("<TABLE BORDERWIDTH=0 WIDTH=\"98%%\">\n", r); 
      ap_rputs("<TR><TD><TABLE ><TR><TD>\n", r);
      ap_rprintf(r, "<TABLE >\n<TR><TD WIDTH=100 ALIGN=RIGHT>%s:</TD><TD>\t", "Reservation number");
      //ap_rprintf(r, "<INPUT TYPE=\"%s\" SIZE=\"%s\" MAXLENGTH=\"%s\" NAME=\"%s\"", "TEXT", 15, 15, "name");
      ap_rprintf(r, "<INPUT TYPE=TEXT ");
 
      ap_rprintf(r, "VALUE=\"%s\" >", "");
      ap_rputs("\t</TD>", r);
      ap_rputs("</TR></TABLE>\n", r); 
      
      ap_rputs("</TD><TD>\n", r);
      ap_rputs("</TD></TR></TABLE></TD></TR>\n",r );
      ap_rputs("</TD></TABLE>\n<INPUT TYPE=SUBMIT VALUE=submit >\n<INPUT TYPE=RESET VALUE=clear >\n</FORM>\n", r); 
      ap_rputs(QOS_FOOTER, r);
      //ap_rputs(QOS_CANCEL_HEADER,r);
      //ap_rputs(QOS_FOOTER, r);
      return OK;
    } 
    else if(!strncmp(r->method, "POST", 4)) {
      char *cid = NULL;

      fprintf(stderr, "cancel post request\n");
      r->content_type = "text/html";
      ap_send_http_header(r);

      if(!cgi_init_post(r, &cid)) {
        ap_table_set(r->subprocess_env, "E2E_TYPE", "cancel");
        ap_table_set(r->subprocess_env, "E2E_HANDLE_ID", "2078917053");
      }
      else { 
        return HTTP_INTERNAL_SERVER_ERROR;
      }
    }
  }

  return DECLINED;
}
 
static handler_rec e2eform_handlers[] = {
  {"*/*", e2eform_handler},
  {NULL}
};
 
module MODULE_VAR_EXPORT e2eform_module = {
  STANDARD_MODULE_STUFF,
  NULL, /* initializer */
  NULL, /* create per-dir config */
  NULL, /* merge per-dir config */
  NULL, /* create per-server config */
  NULL, /* merge per-server config */
  NULL, /* command table */
  e2eform_handlers, /* [#8] handlers */
  NULL, /* [#1] file name translation */
  NULL, /* [#4] check user id */
  NULL, /* [#5] check auth */
  NULL, /* [#3] check access */
  NULL, /* [#6] type checker */
  NULL, /* [#7] prerun fixups */
  NULL, /* [#9] logger */
  NULL, /* [#2] header parser */
  NULL, /* child init */
  NULL, /* child exit */
  NULL, /* [#0] post read request */
};     

char * fix_query(char * query)
{
  int i,o, len;
  char numbuf[3];
 
  len= strlen(query);
  for( i= 0; i< len ; i++) {
    switch(query[i])
    {
    case '+': 
     query[i] = ' ';
    break;
    case '%':
     numbuf[0] = query[i+1];
     numbuf[1] = query[i+2];
     numbuf[2] = '\0';
     query[i] = (char)strtol( numbuf, (char **)NULL, 16); 
     for( o = i+1 ; o < (strlen(query) - 2) ; o++) 
       query[o] = query[o+2];
     len-= 2;
    break;
    }
  }
  return query; 
}

char * cgi_get_var(char * name, char ** post, pool *p)
{
  return cgi_var(name, CGI_GET, post, p);
}

char * cgi_var(char *name, int type, char ** POST, pool *p)
{
  char *query = NULL;
  char *p1, *p2;

  if( type == CGI_GET ) {
    //query = copy_string((char *)getenv("QUERY_STRING"));
    query = (char *)getenv("QUERY_STRING");
    if(query == NULL) return NULL;
  }
  else if( type == CGI_POST ) {
    if( *POST == NULL || (query = copy_string(p, *POST)) == NULL) 
      return NULL;
  }



  p1 = p2 = query;

  while( *p1 != '\0' )
  {
    for(; *p2 != '=' && *p2 != '\0'; p2++);
    if( *p2 == '\0' ) return NULL;
    
    *p2 = '\0'; 

    if(compare_strings(p1, name)) {
      p2++;
      p1 = p2;
      for(; *p2 != '&' && *p2 != '\0'; p2++);
      *p2 = '\0';
      p1 = fix_query(p1);
      return copy_string(p, p1);
    } 
    else {
      p2++;
      for(; *p2 != '&' && *p2 != '\0'; p2++);
      if( *p2 == '&' ){
        p2++;
      }
      p1 = p2;
    }
  }
  return NULL;
}

int cgi_init_post(request_rec *r, char **POST)
{
  char * ret;
  int i, len, retval;
  char *data = NULL;
  array_header *pArray;
  table_entry *pElts;
 
  ap_add_common_vars(r);
  ap_add_cgi_vars(r);
 
  pArray=ap_table_elts(r->subprocess_env);
  pElts=(table_entry *)pArray->elts;
 
  for(i=0 ; i < pArray->nelts ; ++i) {
    if(!strncmp(pElts[i].key, "CONTENT_LENGTH", 14)) {
      fprintf(stderr, "%s %s ", pElts[i].key, pElts[i].val);
      len = atoi(pElts[i].val);
      if((retval = ap_setup_client_block(r, REQUEST_CHUNKED_DECHUNK))
                   != OK) {
        fprintf(stderr, "ap_setup_client_block failed\n");
        return 0;
      }
      if(!ap_should_client_block(r)) {
        fprintf(stderr, "ap_should_client_block failed\n");
        return 0;
      }
      (*POST) = (char *)ap_pcalloc(r->pool, len+1);
      if(ap_get_client_block(r, (*POST), len) < 0) {
        fprintf(stderr, "ap_get_client_block failed\n");
        return 0;
      }
      (*POST)[len] = '\0';
      return 1;	
    }
  }
  return 0;
}

char * inet_addr_check( char * value )
{
  int i;

  for( i=0 ; i < strlen( value ) ; i++ ) {
    if(!( isdigit(value[i]) || isalpha(value[i]) || value[i] == '.' ||
          value[i] == '-' || value[i] == '_'  )) {
      return "value must contain only letters, digits, '.'s, '-'s and '_'s";
    }
  }

  /* checking i+1 .. thus the -1 ! */
  for( i=0 ; i < (strlen( value ) -1) ; i++ ) { 
    if( (value[i] == '.' || value[i] == '-' || value[i] == '_') &&
        (value[i+1] == '.' || value[i+1] == '-' || value[i+1] == '_' ))
      return "punctuation must be separated!";
  }

  /* make sure that punctuation exists!  */
  for( i=0 ; i < strlen( value ) ; i++ ) {
    if( value[i] == '.' ) break;
  }
  if( i == strlen( value ) ){
    return "where is the punctuation?";
  }
 

  return NULL;
}

char * int_check(char * val)
{
  int i;

  for( i=0 ; i < strlen(val) ; i++ )
  {
    if( !isdigit(val[i]) ) {
      return "must only contain digits.";
    }
  }

  return NULL;
}



/* check an INET_4 address  */
char * check_inet_4_addr( char * value )
{
   int i,x,y;
   int len;
   int addr[4];

   len = strlen( value );

   /* inet 4 addresses are in the form of X.X.X.X where X is a number from 0 */
   /*        to 255                                                          */

   /* first check to make sure that _ALL_ chars are digits or '.''s */
   for( i=0 ; i<len ; i++ )
   {
     if( ! (isdigit(value[i]) || value[i] == '.' ) )
       return "not a numeric inet_4 IP addresses";
   }

   /* periods can not be first or last */
   if( value[0] == '.' || value[len] == '.')
     return "Malformed inet_4 IP address.<BR>Should be XXX.XXX.XXX.XXX where 0=< XXX =<255";

   /* count number of periods: */
   x=0; 
   for( i=0 ; i<len ; i++ )
     if( value[i] == '.' ) x++; 
   if( x!=3 ) return "Malformed inet_4 IP address.<BR>Should be XXX.XXX.XXX.XXX where 0=< XXX =<255";

   sscanf(value, "%d.%d.%d.%d", &addr[0],&addr[1],&addr[2],&addr[3]);
                 
   for(i=0 ; i<4 ; i++ ){
     if(addr[i] < 0 || addr[i] > 255 ) 
       return "Malformed inet_4 IP address.<BR>Should be XXX.XXX.XXX.XXX where <B>0=< XXX =<255</B>";
   }


   /* type check passed:  */
   return NULL;
}

/* ERRORS: errors are removed when displayed. */

/* check form.  Make sure that it is properly initialized.	*/
/* returns the number of errors encountered in the form.	*/

int FORM_check(FORM * f, request_rec *r) 
{
  int i = 0;

  if(f->inputs == NULL ){
    ap_rputs("<P>FORM ERROR: no inputs defined</P>\n", r);
    i++;
  }
  if(f->method != CGI_POST && f->method != CGI_GET ){
    ap_rputs("<P>FORM ERROR: unsupported method.\n", r);
    i++;
  }
  if(f->action == NULL ){
    ap_rputs("<P>FORM ERROR: no action defined.</P>\n", r);
    i++;
  }
  return i;
}


/* prints form with all of its inputs, diplays errors (and then destroys them)*/
void FORM_print(FORM * f, request_rec *r)
{
  int i,o;

  if( FORM_check(f, r) != 0 ) {
    return;
  }

 
  ap_rputs("<FORM METHOD=\"", r);
  if( f->method == CGI_GET  )      ap_rputs("GET", r); 
  else if( f->method == CGI_POST ) ap_rputs("POST", r); 

  ap_rprintf(r, "\" ACTION=\"%s\">\n", f->action);

  ap_rputs("<TABLE BORDERWIDTH=0 WIDTH=\"98%%\">\n", r);

  for(i=0; f->inputs[i] != NULL ; i++ ) { 
    ap_rputs("<TR><TD><TABLE ><TR><TD>\n", r);
    for(o=0; f->inputs[i][o] != NULL ; o++ ){
      INPUT_print(f->inputs[i][o], r);
      ap_rputs("</TD><TD>\n", r);
    }
    ap_rputs("</TD></TR></TABLE></TD></TR>\n",r );
  }

  ap_rputs("</TD></TABLE>\n<INPUT TYPE=SUBMIT VALUE=create >\n<INPUT TYPE=RESET  VALUE=clear >\n</FORM>\n", r);
}

void INPUT_add_error( INPUT * inp, char * msg, pool *p)
{
  if( inp->error == NULL )
    inp->error = (char *) copy_string(p, msg);
  else
    sprintf( inp->error, "%s\n%s", inp->error, msg);
}

char * FORM_find_value(FORM *form, char * name)
{
  int i,o;

  for(i=0 ; form->inputs[i] != NULL ; i++ )
  {
  for( o=0 ; form->inputs[i][o] != NULL ; o++ )
  {
    if( compare_strings(form->inputs[i][o]->name, name) )
      return form->inputs[i][o]->value;
  }
  }  
  return NULL;
}

void FORM_error( FORM * form, char* input_name, char * err_msg, pool *p )
{
  int i,o;


  for(i=0 ; form->inputs[i] != NULL ; i++ )
  {
    for(o=0 ; form->inputs[i][o] != NULL ; o++ )
    {
     if( compare_strings(form->inputs[i][o]->name, input_name) ) {
       INPUT_add_error(form->inputs[i][o], err_msg, p);
       return;
     }
    }
  }
       return;
}
 
void INPUT_print( INPUT * input, request_rec *r )
{
  int o;

  if( input == NULL ) return;
  ap_rputs("<TABLE >\n<TR><TD ", r);
  
  if( input->label != NULL && strlen(input->label) != 0 ) {
    ap_rputs("WIDTH=100 ALIGN=RIGHT", r);
  }
  ap_rputs("> ", r);
  
  if( input->label != NULL )
    ap_rprintf(r, "%s:", input->label );
  
  ap_rputs("</TD><TD>\t", r);
  
  /* INPUT */
  if( input->option_list == NULL ) {
    ap_rprintf(r, "<INPUT TYPE=\"%s\" SIZE=\"%s\" MAXLENGTH=\"%s\" NAME=\"%s\"",
	       input->type, input->size, input->max_len,
	       input->name);

    if( input->value != NULL )
      ap_rprintf(r, "VALUE=\"%s\" >", input->value);
    }
    else /* <SELECT... */
    {
      ap_rprintf(r, "<SELECT NAME=\"%s\" SIZE=\"%s\">\n",
		 input->name, input->size);

      for( o=0; input->option_list[o] != NULL ; o++ )
      {
         ap_rprintf(r, "<OPTION VALUE=\"%s\"", input->option_list[o]);

         if( input->value != NULL &&
             compare_strings( input->value,input->option_list[o])
         )
         {
           ap_rputs("SELECTED", r);
         }

         ap_rprintf(r, " >%s</OPTION>\n", input->option_list[o] );
      }

      ap_rputs("</SELECT>\n", r);

    }

    ap_rputs("\t</TD>", r);

    if( input->error != NULL )
    {
      ap_rprintf(r, "<TD VALIGN=\"bottom\"><FONT COLOR=\"red\"><PRE>%s</PRE></FONT></TD>", input->error);
      input->error = NULL;
     }

    ap_rputs("</TR></TABLE>\n", r);
}



INPUT * INPUT_init(char * name, char * label, char* (*f_check)(char *), 
		   char * type, char * size, char *max_len, 
		   char **option_list, pool *p)
{
  INPUT * in = NULL;

  if((in = (INPUT *) ap_pcalloc(p, sizeof(INPUT))) == NULL)
    return NULL;

  if((in->name = copy_string(p, name)) == NULL) return NULL;
  in->label	= copy_string(p, label);
  in->type	= copy_string(p, type);
  if((in->size	= copy_string(p, size)) == NULL) return NULL;
  if((in->max_len = copy_string(p, max_len)) == NULL) return NULL;
  in->option_list = option_list;
  in->f_check	= f_check; 

  return in;
}

FORM * FORM_init(INPUT *** inputs, int method, char * action, pool *p)
{
  FORM * f;

  if((f = (FORM *) ap_pcalloc(p, sizeof(FORM))) == NULL) return NULL;

  if((f->action = copy_string(p, action)) == NULL) return NULL;
  f->inputs = inputs;
  f->method = method;

  return f;
}

int post_form(request_rec *r)
{
  FORM * form;

  if( (form = set_up_form(r->pool)) == NULL )
    return 1;

  r->content_type = "text/html";
  ap_send_http_header(r);
  ap_rputs(QOS_CREATE_HEADER, r); 

  set_date( form, r );

  FORM_print(form, r);
  ap_rputs(QOS_FOOTER, r);

  return 0;
}

int read_form(request_rec *r)
{
  FORM * form = NULL;
  char * ret = NULL;
  char * post = NULL;

  if( (form = set_up_form(r->pool)) == NULL )
    return 1;
 
  if( form->method == CGI_POST ) 
    if(!cgi_init_post(r, &post))
      return 1;

  r->content_type = "text/html";
  ap_send_http_header(r);
  ap_rputs(QOS_CREATE_HEADER, r); 

  if(!setvar(form, post, r) )
  {
    FORM_print(form, r);
    ap_rputs(QOS_FOOTER, r);
    return 1;
  }
  else
  {
    if((ret = send_var(form, r)) != NULL) {
      ap_rprintf(r, "<P><FONT COLOR=red>%s</FONT></P>", ret); 
      FORM_print(form, r);
      ap_rputs(QOS_FOOTER, r);
      return 1;
    }
  }
  
  //ap_rputs(QOS_FOOTER, r);
  return 0;
}

int setvar(FORM * f, char *post, request_rec *r)
{
  int i,o;
  int err = 0;
  char * ret;
  
  for( i=0 ; f->inputs[i] != NULL ; i++ )
  {
  for( o=0 ; f->inputs[i][o] != NULL ; o++ )
  {
      f->inputs[i][o]->value = cgi_var(f->inputs[i][o]->name,f->method, 
	                               &post, r->pool);
        
      if( f->inputs[i][o]->value == NULL || 
          strlen(f->inputs[i][o]->value)==0){ 
        INPUT_add_error(f->inputs[i][o], "error reading value.", r->pool);
        err++;
      } 
      else {
        if( f->inputs[i][o]->f_check != NULL && 
            (ret = f->inputs[i][o]->f_check(f->inputs[i][o]->value)) != NULL ){
          INPUT_add_error( (INPUT *) f->inputs[i][o] , (char *) ret, r->pool);
          err++;
        }
      }
  }
  }
  
  if( !err )   	return 1;
  else 		return 0;
}

FORM * set_up_form(pool *p)
{
  FORM *  f;
  INPUT ***lines;
  int i;
  char ** protocol_options; 
  char ** month_tab;


  if((lines = (INPUT ***) ap_pcalloc(p, sizeof( INPUT **) * 8 )) == NULL)
    return NULL;

  for( i=0 ; i < 8 ; i++ ){
    if((lines[i] = (INPUT **) ap_pcalloc(p, sizeof(INPUT * ) * 4)) == NULL)
      return NULL;
  }

  if((protocol_options = (char **) ap_pcalloc(p, sizeof(char *) * 3)) == NULL)
    return NULL;
  
  if((protocol_options[0] = copy_string(p, "TCP")) == NULL) return NULL;
  if((protocol_options[1] = copy_string(p, "UDP")) == NULL) return NULL;
  protocol_options[2] = NULL;	/*terminate array*/

  if((month_tab = (char ** ) ap_pcalloc(p, sizeof(char *) * 13)) == NULL)
    return NULL;

  if((month_tab[0] = copy_string(p, "JAN")) == NULL) return NULL;
  if((month_tab[1] = copy_string(p, "FEB")) == NULL) return NULL;
  if((month_tab[2] = copy_string(p, "MAR")) == NULL) return NULL;
  if((month_tab[3] = copy_string(p, "APR")) == NULL) return NULL;
  if((month_tab[4] = copy_string(p, "MAY")) == NULL) return NULL;
  if((month_tab[5] = copy_string(p, "JUN")) == NULL) return NULL;
  if((month_tab[6] = copy_string(p, "JUL")) == NULL) return NULL;
  if((month_tab[7] = copy_string(p, "AUG")) == NULL) return NULL;
  if((month_tab[8] = copy_string(p, "SEP")) == NULL) return NULL;
  if((month_tab[9] = copy_string(p, "OCT")) == NULL) return NULL;
  if((month_tab[10] = copy_string(p, "NOV")) == NULL) return NULL;
  if((month_tab[11] = copy_string(p, "DEC")) == NULL) return NULL;
  month_tab[12] = NULL;

  if( (lines[0][0] = INPUT_init("HOST1", "source host", check_inet_4_addr, 
		           "TEXT" , "15", "15", NULL, p)) == NULL ) 
    return NULL;
  lines[0][1] = NULL;

  
  if( (lines[1][0] = INPUT_init("HOST2", "destination host", check_inet_4_addr,
                           "TEXT" , "15", "15", NULL, p)) == NULL ) 
    return NULL;
  lines[1][1] = NULL;


  if( (lines[2][0] = INPUT_init("BANDWIDTH", "bandwidth", int_check, 
                           "TEXT" , "5", "10", NULL, p)) == NULL ) 
    return NULL;
  lines[2][1] = NULL;

  if( (lines[3][0] = INPUT_init("DURATION", "duration",  int_check,
                           "TEXT" , "5", "10", NULL, p)) == NULL ) 
    return NULL;
  lines[3][1] = NULL;
  
  if( (lines[4][0] = INPUT_init("PROTOCOL", "protocol",  NULL,
                           NULL , "1", "", protocol_options, p)) == NULL ) 
    return NULL;
  lines[4][1] = NULL;

  
  if( (lines[5][0] = INPUT_init("DAY", "date",  int_check,
                           "TEXT" , "2", "2", NULL, p)) == NULL ) 
    return NULL;

  if( (lines[5][1] = INPUT_init("MONTH", NULL, NULL,
                           NULL, "1", "", month_tab, p)) == NULL ) 
    return NULL;

  if( (lines[5][2] = INPUT_init("YEAR", NULL, int_check,
                           "TEXT", "4", "4", NULL, p)) == NULL ) 
    return NULL;
  lines[5][3] = NULL;

  if( (lines[6][0] = INPUT_init("HOUR", "time", int_check,
   			   "TEXT", "2", "2", NULL, p)) == NULL ) 
    return NULL;
  if( (lines[6][1] = INPUT_init("MINUTE", NULL, int_check,
   			   "TEXT", "2", "2", NULL, p)) == NULL ) 
    return NULL;
  if( (lines[6][2] = INPUT_init("SECOND", NULL, int_check,
   			   "TEXT", "2", "2", NULL, p)) == NULL ) 
    return NULL;
  lines[6][3] = NULL;

  lines[7] = NULL; 

  if( (f =  FORM_init(lines, CGI_POST, QOS_CREATE_ACTION, p)) == NULL ) 
    return NULL;
 
  return f;
}

char * send_var(FORM * f, request_rec *r)
{
  char *host1, *host2, *duration, *bandwidth, *protocol, *offset; 
  table *t = r->subprocess_env;

  host1=host2=duration=bandwidth=protocol=offset=NULL;

  if((host1     =FORM_find_value(f,"HOST1"    )) == NULL ){
    return "Error: value for \"host1\" not found\n";
  }

  if((host2     =FORM_find_value(f,"HOST2"    )) == NULL ){ 
    return "Error: value for \"host2\" not found\n";
  }

  if((duration  =FORM_find_value(f,"DURATION" )) == NULL ){
    return "Error: value for \"duration\" not found\n";
  }

  if((protocol  =FORM_find_value(f,"PROTOCOL" )) == NULL ){
    return "Error: value for \"protocol\" not found\n";
  }

  if((bandwidth =FORM_find_value(f,"BANDWIDTH")) == NULL ){
    return "Error: value for \"bandwidth\" not found\n";
  }

  if((offset = find_offset(f, r)) == NULL) {
    return "Error calculating offset\n";
  }

  fprintf(stderr, "mod_e2eform: host1 = %s\nhost2 = %s\nbandwidth = %s\noffset = %s\nduration = %s\nprotocol = %s\n", host1, host2, bandwidth, offset, duration, protocol);

  ap_table_set(t, "E2E_HOST1", host1);
  ap_table_set(t, "E2E_HOST2", host2);
  ap_table_set(t, "E2E_OFFSET", offset);
  ap_table_set(t, "E2E_BANDWIDTH", bandwidth);
  ap_table_set(t, "E2E_DURATION", duration);
  ap_table_set(t, "E2E_PROTOCOL", protocol);
  ap_table_set(t, "E2E_TYPE", "create");
  ap_table_set(t, "E2E_HANDLE_ID", "");
 
  ap_rprintf(r, "<P>host1 = %s</P>\n<P>host2 = %s</P>\n<P>duration = %s</P>\n<P>bandwidth = %s</P>\n<P>protocol = %s</P>\n<P>offset = %s</P>\n", host1, host2, duration, bandwidth, protocol, offset);

 return NULL;	/* everything worked */
}

int set_date( FORM * form, request_rec *r )
{
   int i, o;
   struct tm * tm;
   char tmp[100];
   time_t curr_time;

   if( (curr_time = (time_t)time((time_t)NULL)) == (time_t) -1 ) {
     ap_rputs("<P>Error reading current time</P>\n", r);
     return 0;
   }

   if((tm = localtime(&curr_time)) == NULL) {
     ap_rputs("<P>Error reading current time\n</P>", r);
     return 0;
   }

   for(i=0 ; form->inputs[i]    != NULL ; i++){
   for(o=0 ; form->inputs[i][o] != NULL ; o++){
     if( compare_strings("HOUR", form->inputs[i][o]->name ) ){
       if( snprintf(tmp, 100, "%d", tm->tm_hour) != 0 ){
         form->inputs[i][o]->value = copy_string(r->pool, tmp);  
       }else{ printf("error reading value\n"); return 0; }
     }
     else if( compare_strings("MINUTE", form->inputs[i][o]->name ) ){
       if( snprintf(tmp, 100, "%d", tm->tm_min) != 0 ){
         form->inputs[i][o]->value = copy_string(r->pool, tmp);
       }else{ printf("error reading value\n"); return 0;}
     }     
     else if( compare_strings("SECOND", form->inputs[i][o]->name ) ){
       if( snprintf(tmp, 100, "%d", tm->tm_sec) != 0 ){
         form->inputs[i][o]->value = copy_string(r->pool, tmp);
       }else{ printf("error reading value\n"); return 0;}
     }     
     else if( compare_strings("DAY", form->inputs[i][o]->name ) ){
       if( snprintf(tmp, 100, "%d", tm->tm_mday) != 0 ){
         form->inputs[i][o]->value = copy_string(r->pool, tmp);
       }else{ printf("error reading value\n"); return 0;}
     }     
     else if( compare_strings("MONTH", form->inputs[i][o]->name ) ){
       if( tm->tm_mon < 12 && tm->tm_mon >= 0  ){
         form->inputs[i][o]->value = copy_string(r->pool, monthtt[tm->tm_mon]);
       }else{ printf("error reading value\n"); return 0;}
     }     
     else if( compare_strings("YEAR", form->inputs[i][o]->name ) ){
       if( snprintf(tmp, 100, "%d", tm->tm_year + 1900 ) != 0 ){
         form->inputs[i][o]->value = copy_string(r->pool, tmp);
       }else{ printf("error reading value\n"); return 0;}
     }     

   }
   }
   return 1;
}

char * find_offset(FORM * form, request_rec *r)
{
  char * offset;
  int ofs;
  
  time_t now;
  int i;
  char * errcheck;

  char *hour=NULL, *min=NULL, *sec=NULL, *day=NULL, *mon=NULL, *year=NULL;
  int      h,   mi,    s,    d,   mo=-1,     y;

  if((hour = FORM_find_value(form, "HOUR")) == NULL) return NULL;
  if((min  = FORM_find_value(form, "MINUTE")) == NULL) return NULL;
  if((sec  = FORM_find_value(form, "SECOND")) == NULL) return NULL;
  if((day  = FORM_find_value(form, "DAY")) == NULL) return NULL;
  if((mon  = FORM_find_value(form, "MONTH")) == NULL) return NULL;
  if((year = FORM_find_value(form, "YEAR")) == NULL) return NULL;

  h  = strtol( hour, &errcheck, 10 );  
  if( *errcheck != '\0' ) return NULL;
  
  mi = strtol( min , &errcheck, 10 );
  if( *errcheck != '\0' ) return NULL;

  s  = strtol( sec , &errcheck, 10 );
  if( *errcheck != '\0' ) return NULL;

  d  = strtol( day , &errcheck, 10 );
  if( *errcheck != '\0' ) return NULL;

  y  = strtol( year,  &errcheck, 10 );
  if( *errcheck != '\0' ) return NULL;

  if(errno == ERANGE) { 
    return NULL;
  }

  y -=1900;	/* year adjustment */

  for(i=0; i<12 ; i++ ) {
    if( compare_strings(mon, monthtt[i]) ){
      mo = i;
      break;
    }
  }
  if(i==12) return NULL;

   if( (now = (time_t)time((time_t)NULL)) == (time_t) -1 ) {
     ap_rputs("<P>Error reading current time</P>\n", r);
     return 0;                                     
   }

   offset = (char *) ap_pcalloc(r->pool, sizeof(char) * 100 );
   ofs = sec_offset(&now,s,mi,h,d,mo,y);
   if( ofs < 0 ) {
    ap_rputs("<P>Requested activation time has allready passed.</P>\n", r);
    return NULL;
   }

   snprintf(offset, 100, "%d", ofs ); 
 
   return offset; 
}

int sec_offset( time_t * now, int sec, int min, int hour, int day, 
	        int month, int year)
{
  time_t then;
  struct tm then_tm;
  double d_offset;

  /* range checking done by mktime */
  then_tm.tm_sec = sec;
  then_tm.tm_min = min;
  then_tm.tm_hour = hour;
  then_tm.tm_mday = day;
  then_tm.tm_mon  = month;
  then_tm.tm_year = year;
  then_tm.tm_isdst = -1;

  then = mktime( &then_tm );

  if( then == (time_t)-1 ) {
    return -1;              
  }
   
  d_offset = difftime(then, *now);

  return (int) d_offset;
}

char * copy_string(pool *p, char * src)
{
  int len = 0;
  char *dest = NULL;

  if(src == NULL) return NULL;
  len = strlen(src);
  
  if((dest = (char *) ap_pcalloc(p, sizeof(char) * (len+1) )) == NULL)
    return NULL;

  strncpy(dest,src,len);

  return dest;
}
 
int compare_strings(char * one, char * two)  /* 1 if same, 0 if fail */
{
  if( one == NULL || two == NULL ) return 0;
  if( strlen(one) != strlen(two) ) return 0;

  if( strncmp(one, two, strlen(one)) == 0 ) return 1;
  else return 0;
}

void print_file( char * filename ) 
{
  FILE * fd;

  int ret;
  char c;

  fd = fopen(filename, "r");
  if( fd == NULL ) {
    printf("<P>Error opening file: %s</P>\n", filename); 
    return;
  }

  while( (ret = fscanf(fd, "%c", &c )) != EOF )
    printf("%c", c);

  fclose(fd);
}
